--  SPDX-License-Identifier: GPL-2.0+
--  Avalon-ST to SpaceWire light interface
--   Copyright 2020 Max Henkel <henkel@tugraz.at>

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity avlstspw is

    port (
        -- System clock.
        clk:        in  std_logic;

        -- Synchronous reset (active-low).
        rstn:       in  std_logic;

	-- Avalon ST input
        asi_data       : in  std_logic_vector(7 downto 0);
        asi_valid      : in  std_logic;
        asi_ready      : out std_logic                       := 'X';
        asi_sop        : in  std_logic;
        asi_eop        : in  std_logic;
        --asi_empty      : in  std_logic_vector(0 downto 0);
        --asi_error      : in  std_logic_vector(0 downto 0);


        -- Request transmission of an N-Char.
        -- Keep this signal high until confirmed by "txack".
        txwrite:    out std_logic                            := 'X';

        -- Control flag to be sent with the next N-Char.
        -- Must be valid while "txwrite" is high.
        txflag:     out std_logic := 'X';

        -- Byte to send, or "00000000" for EOP or "00000001" for EEP.
        -- Must be valid while "txwrite" is high.
        txdata:     out std_logic_vector(7 downto 0)         := (others => 'X');

        -- High to confirm transmission of an N-Char.
        -- This is a Wishbone-style handshaking signal; it is combinatorially
        -- dependent on both "fct_in" and "txwrite".
        txack:      in std_logic

    );

end entity avlstspw;

architecture avlstspw_arch of avlstspw is

    signal txeop: std_logic := '0';
    signal txrdy: std_logic := '1'; -- transmitter state

begin
    asi_ready <= txrdy and not txeop;

    -- can be optimized because of Wishbone-style handshake (txack)
    process (clk) is
    begin
        if rising_edge(clk) then
            -- Synchronous reset
            if rstn = '0' then
                txeop <= '0';
                txrdy <= '1';
                txwrite <= '0';
            elsif txrdy = '1' then
                -- either tx EOP marker
                if txeop = '1' then
                    txwrite <= '1';

                    txflag <= '1';
                    txdata <= "00000000";

                    txeop <= '0';
                    txrdy <= '0';
                -- or take new input from master
                elsif asi_valid = '1' then
                    -- tx EOP in next cycle
                    if asi_eop = '1' then
                        txeop <= '1';
                    end if;

                    txwrite <= '1';
                    txflag <= '0';
                    txdata <= asi_data;
                    -- let master wait until symbol is finished
                    txrdy <= '0';
                end if;
            end if;

            if txack = '1' then
                txrdy <= '1';
                txwrite <= '0';
                txdata <= "00000000";
                txflag <= '0';
            end if;
        end if;
    end process;

end architecture avlstspw_arch;
