--  SPDX-License-Identifier: GPL-2.0+
--  SpaceWire core with Avalon-MM/ST interface by M. Henkel <henkel@tugraz.at>
--   adapted from SpaceWire core with AMBA interface (spwamba.vhd)
--   by Joris van Rantwijk <joris@jorisvr.nl>
--      https://opencores.org/projects/spacewire_light
--
--------------------------------------------------------------------------
--
--  Avalon-MM registers:
--
--      Address 0x00: Control Register
--          bit 0     Reset spwavl core (auto-clear)
--          bit 1     Reserved
--          bit 2     Link start
--          bit 3     Link autostart
--          bit 4     Link disable
--          bit 5     Enable timecode transmission through tick_in signal
--          bit 6:8   Reserved
--          bit 9     Enable interrupt on link up/down
--          bit 10    Enable interrupt on time code received
--          bit 11:12 Reserved
--          bit 13    Enable interrupt on RX packet
--          bit 14    Enable interrupt on error
--          bit 27:24 Reserved
--
--      Address 0x04: Status Register
--          bit 1:0   Link status: 0=off, 1=started, 2=connecting, 3=run
--          bit 2     Got disconnect error (sticky)
--          bit 3     Got parity error (sticky)
--          bit 4     Got escape error (sticky)
--          bit 5     Got credit error (sticky)
--          bit 6:9   Reserved
--          bit 10    Received timecode (sticky)
--          bit 11:12 Reserved
--          bit 13    Received packet (sticky)
--          bit 14    RX buffer empty after packet
--
--          Sticky bits are reset by writing a '1' bit to the corresponding
--          bit position(s).
--
--      Address 0x08: Transmission Clock Scaler
--          bit 7:0   txclk division factor minus 1
--          bit 12:8  data line delay in clock cycles
--          bti 14:13 reserved
--          bit 15    data line inversion
--          bit 20:16 strobe line delay in clock cycles
--          bit 21    reserved
--          bit 22    data / strobe line swap
--          bit 23    strobe line inversion
--
--      Address 0x0c: Timecode Register
--          bit 5:0   Last received timecode value (read-only)
--          bit 7:6   Control bits received with last timecode (read-only)
--          bit 13:8  Timecode value to send on next tick_in (auto-increment)
--          bit 15:14 Reserved (write as zero)
--          bit 16    Write '1' to send a timecode (auto-clear)
--
--      Address 0x1c: Version Register (read-only)
--          bit 0     timecode function enabled
--          bit 6     rximpl (0 -> generic, 1 -> fast)
--          bit 7     tximpl (0 -> generic, 1 -> fast)
--          bit 11:8  rx FIFO size as log2
--          bit 15:12 tx FIFO size as log2
--          bit 23:16 system clock in MHz
--          bit 27:24 minor version
--          bit 31:28 major version
--
--------------------------------------------------------------------------
--
--  v0.1
--    - initial adaptation
--
--  v0.2
--    - version register added
--    - interrupt generation improved (more IE reasons)
--    - fixed tx buffer issue with lost byte at the end
--    - fixed infinite zero packets, when asi_eop is held high without asi_valid
--
--  v0.3
--    - reworked tx buffer to a simple converter stage without internal buffer
--
--------------------------------------------------------------------------
--  Known issues / TODO
--    - handling of EEP as aso_error for rx path (and possibly tx)
--    - correct rx buffer handling to remove '0x00' byte at end of packet
--    - refactor and reimplement rx buffer with e.g. 32 bit width
--      (partly done) and dual clocks.
--    - txdiscard for tx chain
--    - set timing constraints in Quartus/TimeQuest

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.spwpkg.all;

entity spwavl is

    generic (
        SPW_VER_MAJOR : integer range 0 to 15 := 0;
        SPW_VER_MINOR : integer range 0 to 15 := 3;

        -- System clock frequency in Hz.
        -- This must be set to the frequency of "clk". It is used to setup
        -- counters for reset timing, disconnect timeout and to transmit
        -- at 10 Mbit/s during the link handshake.
        sysfreq:        integer := 10000000;

        -- Transmit clock frequency in Hz (only if tximpl = impl_fast).
        -- This must be set to the frequency of "txclk". It is used to
        -- transmit at 10 Mbit/s during the link handshake.
        txclkfreq:      integer := 0;

        -- Selection of a receiver front-end implementation.
        rximpl:         spw_implementation_type := impl_generic;

        -- Maximum number of bits received per system clock
        -- (must be 1 in case of impl_generic).
        rxchunk:        integer range 1 to 4 := 1;

        -- Selection of a transmitter implementation.
        tximpl:         spw_implementation_type := impl_generic;

        -- Enable capability to generate time-codes.
        timecodegen:    boolean := true;

        -- Size of the receive FIFO as the 2-logarithm of the number of bytes.
        -- Must be at least 6 (64 words = 256 bytes).
        rxfifosize_bits: integer range 6 to 14 := 11;

        txfifosize_bits: integer range 0 to 14 := 0

    );

    port (
        -- System clock.
        clk:        in  std_logic;

        -- Receiver sample clock (only for impl_fast)
        rxclk:      in  std_logic;

        -- Transmit clock (only for impl_fast)
        txclk:      in  std_logic;

        -- Synchronous reset (active-low).
        rstn:       in  std_logic;


	-- Avalon memory-mapped slave
        avs_address    : in  std_logic_vector(2 downto 0)  := (others => 'X'); -- address
        --avs_byteenable : in  std_logic_vector(3 downto 0)  := (others => 'X'); -- byteenable
        avs_write      : in  std_logic                     := 'X';             -- write
        avs_writedata  : in  std_logic_vector(31 downto 0) := (others => 'X'); -- writedata
        avs_read       : in  std_logic                     := 'X';             -- read
        avs_readdata   : out std_logic_vector(31 downto 0);                    -- readdata

        avs_irq        : out std_logic := '0';

	-- Avalon ST output
        aso_data       : out std_logic_vector(7 downto 0) := (others => 'X'); -- 
        aso_valid      : out std_logic                     := 'X';             --
        aso_ready      : in  std_logic                     := 'X';             --
        aso_sop        : out std_logic                     := 'X';             -- 
        aso_eop        : out std_logic                     := 'X';             --
--        aso_empty      : out std_logic_vector(-1 downto 0)  := (others => '0');             --
--        aso_error      : out std_logic_vector(-1 downto 0)  := (others => '0');             --
--        aso_channel    : out std_logic_vector(9 downto 0)  := (others => 'X'); --
	
	-- Avalon ST input
        asi_data       : in  std_logic_vector(7 downto 0) := (others => 'X'); -- 
        asi_valid      : in  std_logic                     := 'X';             -- 
        asi_ready      : out std_logic                     := 'X';             -- 
        asi_sop        : in  std_logic                     := 'X';             --
        asi_eop        : in  std_logic                     := 'X';             -- 
        --asi_empty      : in  std_logic_vector(0 downto 0);                    -- 
        --asi_error      : in  std_logic_vector(0 downto 0);                     --

        -- Pulse for TimeCode generation.
        tick_in:    in  std_logic;

        -- High for one clock cycle if a TimeCode was just received.
        tick_out:   out std_logic;

        -- Data In signal from SpaceWire bus.
        spw_di:     in  std_logic;

        -- Strobe In signal from SpaceWire bus.
        spw_si:     in  std_logic;

        -- Data Out signal to SpaceWire bus.
        spw_do:     out std_logic;

        -- Strobe Out signal to SpaceWire bus.
        spw_so:     out std_logic
    );

end entity spwavl;

architecture spwavl_arch of spwavl is

    -- Convert boolean to std_logic.
    type bool_to_logic_type is array(boolean) of std_ulogic;
    constant bool_to_logic: bool_to_logic_type := (false => '0', true => '1');

    -- Reset time (6.4 us) in system clocks
    constant reset_time:        integer := sysfreq / 156250; -- 1/6.4us

    -- Disconnect time (850 ns) in system clocks
    constant disconnect_time:   integer := sysfreq / 1176470; -- 1/850ns

    -- Initial tx clock scaler (10 Mbit).
    type impl_to_real_type is array(spw_implementation_type) of integer;
    constant tximpl_to_txclk_freq: impl_to_real_type :=
        (impl_generic => sysfreq, impl_fast => txclkfreq);
--    constant effective_txclk_freq: real := tximpl_to_txclk_freq(tximpl);
    constant effective_txclk_freq: integer := tximpl_to_txclk_freq(impl_generic);
    constant default_divcnt:    std_logic_vector(7 downto 0) :=
        std_logic_vector(to_unsigned(integer(effective_txclk_freq / 10e6 - 1), 8));

    type impl_to_logic_type is array(spw_implementation_type) of std_ulogic;
    constant impl_to_logic: impl_to_logic_type := (impl_generic => '0', impl_fast => '1');

    constant SPW_VER : std_logic_vector(7 downto 0) :=
        std_logic_vector( to_unsigned(SPW_VER_MAJOR,4) & to_unsigned(SPW_VER_MINOR,4) );
    constant SPW_CFG : std_logic_vector(23 downto 0) := (
        std_logic_vector( to_unsigned( sysfreq/1e6, 8) ) &      -- 23 downto 16
        std_logic_vector( to_unsigned( txfifosize_bits, 4) ) &  -- 15 downto 12
        std_logic_vector( to_unsigned( rxfifosize_bits, 4) ) &  -- 11 downto  8
        impl_to_logic(tximpl) &      -- tx_impl generic         --  7
        impl_to_logic(rximpl) &      -- rx_impl generic         --  6
        "00000" &                    -- unused                  --  5 downto  1
        bool_to_logic(timecodegen)                              --  0
    );



   component avlstspw is
    port (
        -- System clock.
        clk:        in  std_logic;

        -- Synchronous reset (active-low).
        rstn:       in  std_logic;

	-- Avalon ST input
        asi_data       : in  std_logic_vector(7 downto 0) := (others => 'X'); -- 
        asi_valid      : in  std_logic                     := 'X';             -- 
        asi_ready      : out std_logic                     := 'X';             -- 
        asi_sop        : in  std_logic                     := 'X';             --
        asi_eop        : in  std_logic                     := 'X';             -- 
        --asi_empty      : in  std_logic_vector(0 downto 0);                    -- 
        --asi_error      : in  std_logic_vector(0 downto 0);                     --


        -- Request transmission of an N-Char.
        -- Keep this signal high until confirmed by "txack".
        txwrite:    out std_logic;

        -- Control flag to be sent with the next N-Char.
        -- Must be valid while "txwrite" is high.
        txflag:     out std_logic;

        -- Byte to send, or "00000000" for EOP or "00000001" for EEP.
        -- Must be valid while "txwrite" is high.
        txdata:     out std_logic_vector(7 downto 0);

        -- High to confirm transmission of an N-Char.
        -- This is a Wishbone-style handshaking signal; it is combinatorially
        -- dependent on both "fct_in" and "txwrite".
        txack:      in std_logic

    );
    end component avlstspw;

    -- Registers.
    type regs_type is record
        -- packet state
        rxpacket:       std_logic;      -- '1' when receiving a packet
        rxeep:          std_logic;      -- '1' when rx EEP character pending
        rxeop:          std_logic;      -- '1' when rx EOP character pending
        txpacket:       std_logic;      -- '1' when transmitting a packet
        txdiscard:      std_logic;      -- '1' when discarding a tx packet
        -- RX fifo state
        rxfifo_raddr:   std_logic_vector(rxfifosize_bits-1 downto 0);
        rxfifo_waddr:   std_logic_vector(rxfifosize_bits-1 downto 0);
        rxfifo_wdata:   std_logic_vector(9 downto 0);
        rxfifo_write:   std_ulogic;
        rxfifo_empty:   std_ulogic;
        rxfifo_bytemsk: std_logic_vector(2 downto 0);
        -- FIFO state
        rxfifo_rvalid:  std_logic;      -- '1' if s_rxfifo_rdata is valid
        rxfull:         std_logic;      -- '1' if RX fifo is full
        rxhalff:        std_logic;      -- '1' if RX fifo is at least half full
        rxroom:         std_logic_vector(5 downto 0);

        -- registers
        ctl_reset:      std_ulogic;
        ctl_linkstart:  std_ulogic;
        ctl_autostart:  std_ulogic;
        ctl_linkdis:    std_ulogic;
        ctl_ticken:     std_ulogic;
        ctl_rxstart:    std_ulogic;
        ctl_txstart:    std_ulogic;
        ctl_txcancel:   std_ulogic;
        ctl_ielink:     std_ulogic;
        ctl_ietick:     std_ulogic;
        ctl_ierxdesc:   std_ulogic;
        ctl_ietxdesc:   std_ulogic;
        ctl_ierxpacket: std_ulogic;
        ctl_ieerror:    std_ulogic;
        sta_link:       std_logic_vector(1 downto 0);
        sta_errdisc:    std_ulogic;
        sta_errpar:     std_ulogic;
        sta_erresc:     std_ulogic;
        sta_errcred:    std_ulogic;
        sta_gottick:    std_ulogic;
        sta_rxdesc:     std_ulogic;
        sta_txdesc:     std_ulogic;
        sta_rxpacket:   std_ulogic;
        sta_rxempty:    std_ulogic;
        txdivcnt:       std_logic_vector(7 downto 0);
        tx_delay_data:  std_logic_vector(7 downto 0);
        tx_delay_strobe:std_logic_vector(7 downto 0);
        time_in:        std_logic_vector(5 downto 0);
        tick_in:        std_ulogic;
        -- APB interrupt request
        irq:            std_ulogic;
    end record;

    constant regs_reset: regs_type := (
        rxpacket        => '0',
        rxeep           => '0',
        rxeop           => '1',
        txpacket        => '0',
        txdiscard       => '0',
        rxfifo_raddr    => (others => '0'),
        rxfifo_waddr    => (others => '0'),
        rxfifo_wdata    => (others => '0'),
        rxfifo_write    => '0',
        rxfifo_empty    => '1',
        rxfifo_rvalid   => '0',
        rxfifo_bytemsk  => "000", -- was "111" -> disable bytemask
        rxroom          => (others => '1'),
        rxfull          => '0',
        rxhalff         => '0',
        ctl_reset       => '0',
        ctl_linkstart   => '0',
        ctl_autostart   => '0',
        ctl_linkdis     => '0',
        ctl_ticken      => '0',
        ctl_rxstart     => '0',
        ctl_txstart     => '0',
        ctl_txcancel    => '0',
        ctl_ielink      => '0',
        ctl_ietick      => '0',
        ctl_ierxdesc    => '0',
        ctl_ietxdesc    => '0',
        ctl_ierxpacket  => '0',
        ctl_ieerror     => '0',
        sta_link        => "00",
        sta_errdisc     => '0',
        sta_errpar      => '0',
        sta_erresc      => '0',
        sta_errcred     => '0',
        sta_gottick     => '0',
        sta_rxdesc      => '0',
        sta_txdesc      => '0',
        sta_rxpacket    => '0',
        sta_rxempty     => '1',
        txdivcnt        => default_divcnt,
        tx_delay_data   => (others => '0'),
        tx_delay_strobe => (others => '0'),
        time_in         => (others => '0'),
        tick_in         => '0',
        irq             => '0' );

    signal r: regs_type := regs_reset;
    signal rin: regs_type;

    -- Component interface signals.
    signal recv_rxen:       std_logic;
    signal recvo:           spw_recv_out_type;
    signal recv_inact:      std_logic;
    signal recv_inbvalid:   std_logic;
    signal recv_inbits:     std_logic_vector(rxchunk-1 downto 0);
    signal xmit_rst:        std_logic;
    signal xmiti:           spw_xmit_in_type;
    signal xmito:           spw_xmit_out_type;
    signal xmit_divcnt:     std_logic_vector(7 downto 0);
    signal xmit_delay_data: std_logic_vector(7 downto 0);
    signal xmit_delay_strobe: std_logic_vector(7 downto 0);
    signal link_rst:        std_logic;
    signal linki:           spw_link_in_type;
    signal linko:           spw_link_out_type;

    -- Memory interface signals.
    signal s_rxfifo_raddr:  std_logic_vector(rxfifosize_bits-1 downto 0);
    signal s_rxfifo_rdata:  std_logic_vector(9 downto 0);
    signal s_rxfifo_wen:    std_logic;
    signal s_rxfifo_waddr:  std_logic_vector(rxfifosize_bits-1 downto 0);
    signal s_rxfifo_wdata:  std_logic_vector(9 downto 0);

begin

    -- Instantiate link controller.
    link_inst: spwlink
        generic map (
            reset_time  => reset_time )
        port map (
            clk         => clk,
            rst         => link_rst,
            linki       => linki,
            linko       => linko,
            rxen        => recv_rxen,
            recvo       => recvo,
            xmiti       => xmiti,
            xmito       => xmito );

    -- Instantiate receiver.
    recv_inst: spwrecv
        generic map(
            disconnect_time => disconnect_time,
            rxchunk     => rxchunk )
        port map (
            clk         => clk,

            rxen        => recv_rxen,
            recvo       => recvo,
            inact       => recv_inact,
            inbvalid    => recv_inbvalid,
            inbits      => recv_inbits );

    -- Instantiate receiver front-end.
    recvfront_sel0: if rximpl = impl_generic generate
        recvfront_generic_inst: spwrecvfront_generic
            port map (
                clk         => clk,
                rxen        => recv_rxen,
                inact       => recv_inact,
                inbvalid    => recv_inbvalid,
                inbits      => recv_inbits,
                spw_di      => spw_di,
                spw_si      => spw_si );
    end generate;
    recvfront_sel1: if rximpl = impl_fast generate
        recvfront_fast_inst: spwrecvfront_fast
            generic map (
                rxchunk     => rxchunk )
            port map (
                clk         => clk,
                rxclk       => rxclk,
                rxen        => recv_rxen,
                inact       => recv_inact,
                inbvalid    => recv_inbvalid,
                inbits      => recv_inbits,
                spw_di      => spw_di,
                spw_si      => spw_si );
    end generate;

    -- Instantiate transmitter.
    xmit_sel0: if tximpl = impl_generic generate
        xmit_inst: spwxmit
            port map (
                clk     => clk,
                rst     => xmit_rst,
                divcnt  => xmit_divcnt,
                delay_data => xmit_delay_data,
                delay_strobe => xmit_delay_strobe,
                xmiti   => xmiti,
                xmito   => xmito,
                spw_do  => spw_do,
                spw_so  => spw_so );
    end generate;
    xmit_sel1: if tximpl = impl_fast generate
        xmit_fast_inst: spwxmit_fast
            port map (
                clk     => clk,
                txclk   => txclk,
                rst     => xmit_rst,
                divcnt  => xmit_divcnt,
                xmiti   => xmiti,
                xmito   => xmito,
                spw_do  => spw_do,
                spw_so  => spw_so );
    end generate;

    -- Instantiate RX memory.
    rxmem: spwram
        generic map (
            abits       => rxfifosize_bits,
            dbits       => 10 )
        port map (
            rclk        => clk,
            wclk        => clk,
            ren         => '1',
            raddr       => s_rxfifo_raddr,
            rdata       => s_rxfifo_rdata,
            wen         => s_rxfifo_wen,
            waddr       => s_rxfifo_waddr,
            wdata       => s_rxfifo_wdata );

    -- Instantiate Avalon-ST to SpW link converter
    txavl: avlstspw
        --generic map ( )
        port map (
            clk         => clk,
            rstn        => rstn,
            asi_data    => asi_data,
            asi_valid   => asi_valid,
            asi_ready   => asi_ready,
            asi_sop     => asi_sop,
            asi_eop     => asi_eop,
            txwrite     => linki.txwrite,
            txflag      => linki.txflag,
            txdata      => linki.txdata,
            txack       => linko.txack
        );

    -- Combinatorial process
    process (r, linko, s_rxfifo_rdata, rstn, tick_in,
                aso_ready,
		avs_address, avs_read, avs_write, avs_writedata)  is
        variable v:             regs_type;
        variable v_tmprxroom:   unsigned(rxfifosize_bits-1 downto 0);
        variable v_prdata:      std_logic_vector(31 downto 0);
    begin
        v           := r;
        v_tmprxroom := to_unsigned(0, v_tmprxroom'length);
        v_prdata    := (others => '0');

        -- Reset auto-clearing registers.
        v.ctl_reset     := '0';
        v.ctl_rxstart   := '0';
        v.ctl_txstart   := '0';

        -- Register external timecode trigger (if enabled).
        if timecodegen and r.ctl_ticken = '1' then
            v.tick_in   := tick_in;
        else
            v.tick_in   := '0';
        end if;

        -- Auto-increment timecode counter.
        if r.tick_in = '1' then
            v.time_in   := std_logic_vector(unsigned(r.time_in) + 1);
        end if;

        -- Keep track of whether we are sending and/or receiving a packet.
        if linko.rxchar = '1' then
            -- got character
            v.rxpacket  := not linko.rxflag;
        end if;

        v.rxeop := r.rxpacket and not v.rxpacket;

        v.rxfifo_write := '0';
        if r.rxeep = '1' then -- TODO handle EEP with e.g. Avalon-ST error line
            v.rxfifo_wdata(8)  := '1';
            v.rxfifo_wdata(7 downto 0) := "00000001";
        else
            v.rxfifo_wdata(9)  := ( not r.rxpacket and v.rxpacket );
            v.rxfifo_wdata(8)  := linko.rxflag;
            v.rxfifo_wdata(7 downto 0) := linko.rxdata;
        end if;

        if linko.rxchar = '1' or (r.rxeep = '1' and unsigned(r.rxroom) /= 0) then
            v.rxeep     := '0';
            if r.rxfifo_bytemsk(0) = '0' or linko.rxflag = '1' or r.rxeep = '1' then
--            if linko.rxflag = '1' or r.rxeep = '1' then
                -- Flush the current word to the FIFO.
                v.rxfifo_write   := '1';
                --v.rxfifo_bytemsk := "111";
                v.rxfifo_bytemsk := "000";
            else
                -- Store one byte.
                v.rxfifo_bytemsk := '0' & r.rxfifo_bytemsk(2 downto 1);
            end if;
        end if;

        -- Update RX fifo pointers.
        if (aso_ready = '1') and (r.rxfifo_rvalid = '1') then
            -- read from fifo
            v.rxfifo_raddr  := std_logic_vector(unsigned(r.rxfifo_raddr) + 1);
        end if;
   --     if r.rxfull = '0' then
   --         if (linko.rxchar = '1') or (r.rxeep = '1') then
                -- write to fifo (received char or pending EEP)
        if r.rxfifo_write = '1' then
                v.rxfifo_waddr  := std_logic_vector(unsigned(r.rxfifo_waddr) + 1);
   --         end if;
   --         v.rxeep         := '0';
        end if;

        -- Detect RX fifo empty (using new value of rxfifo_raddr).
        -- Note: The FIFO is empty if head and tail pointer are equal.
        v.rxfifo_empty  := bool_to_logic(v.rxfifo_raddr = r.rxfifo_waddr);
        -- Indicate RX fifo room for SpaceWire flow control.
        -- The flow control window is normally expressed as a number of bytes,
        -- but we don't know how many bytes we can fit in each word because
        -- some words are only partially used. So we report FIFO room as if
        -- each FIFO word can hold only one byte, which is an overly
        -- pessimistic estimate.
        -- (Use the new value of rxfifo_waddr.)
        v_tmprxroom     := unsigned(r.rxfifo_raddr) - unsigned(v.rxfifo_waddr) - 1;
        if v_tmprxroom > 63 then
            -- at least 64 bytes room.
            v.rxroom    := "111111";
        else
            -- less than 64 bytes room.
            -- If linko.rxchar = '1', decrease rxroom by one to account for
            -- the pipeline delay through r.rxfifo_write.
            v.rxroom    := std_logic_vector(v_tmprxroom(5 downto 0) - 
                             unsigned'('0'&linko.rxchar));
        end if;

        -- Keep track of whether the RX fifo contains valid data.
        -- (use new value of rxfifo_raddr)
        v.rxfifo_rvalid := bool_to_logic(v.rxfifo_raddr /= r.rxfifo_waddr);

        -- Update room in RX fifo (use new value of rxfifo_waddr).
        v_tmprxroom := unsigned(r.rxfifo_raddr) - unsigned(v.rxfifo_waddr) - 1;
        v.rxfull    := bool_to_logic(v_tmprxroom = 0);
        v.rxhalff   := not v_tmprxroom(v_tmprxroom'high);
        if v_tmprxroom > 63 then
            v.rxroom    := (others => '1');
        else
            v.rxroom    := std_logic_vector(v_tmprxroom(5 downto 0));
        end if;

        -- If the link is lost, set a flag to discard the current packet.
        if linko.running = '0' then
            v.rxeep     := v.rxeep or v.rxpacket;       -- use new value of rxpacket
--            v.txdiscard := v.txdiscard or v.txpacket;   -- use new value of txpacket
            v.rxpacket  := '0';
--            v.txpacket  := '0';
        end if;

        -- Clear the discard flag when the link is explicitly disabled.
        if r.ctl_linkdis = '1' then
            v.txdiscard := '0';
        end if;

        -- Update status registers.
        v.sta_link(0)   := linko.running or linko.started;
        v.sta_link(1)   := linko.running or linko.connecting;
        if linko.errdisc  = '1' then v.sta_errdisc := '1'; end if;
        if linko.errpar   = '1' then v.sta_errpar  := '1'; end if;
        if linko.erresc   = '1' then v.sta_erresc  := '1'; end if;
        if linko.errcred  = '1' then v.sta_errcred := '1'; end if;
        if linko.tick_out = '1' then v.sta_gottick := '1'; end if;
        if r.rxeop        = '1' then v.sta_rxpacket:= '1'; end if;
        if r.rxfifo_empty = '1' then
            v.sta_rxempty   := '1';
        elsif r.rxfifo_empty = '0' then
            v.sta_rxempty   := '0';
        end if;

        v.irq       :=
            (r.ctl_ielink     and (linko.running xor (r.sta_link(0) and r.sta_link(1)))) or
            (r.ctl_ietick     and linko.tick_out) or
            (r.ctl_ierxpacket and r.rxeop ) or
            (r.ctl_ieerror    and (linko.errdisc or linko.errpar or linko.erresc or linko.errcred));

        -- Avalon-MM read access.
        if avs_read = '1' then
            case avs_address(2 downto 0) is
                when "000" =>   -- read control register
                    v_prdata(0)     := '0';
                    v_prdata(1)     := '0';
                    v_prdata(2)     := r.ctl_linkstart;
                    v_prdata(3)     := r.ctl_autostart;
                    v_prdata(4)     := r.ctl_linkdis;
                    v_prdata(5)     := r.ctl_ticken;
                    v_prdata(6)     := '0';
                    v_prdata(7)     := '0';
                    v_prdata(8)     := r.ctl_txcancel;
                    v_prdata(9)     := r.ctl_ielink;
                    v_prdata(10)    := r.ctl_ietick;
                    v_prdata(11)    := r.ctl_ierxdesc;
                    v_prdata(12)    := r.ctl_ietxdesc;
                    v_prdata(13)    := r.ctl_ierxpacket;
                    v_prdata(14)    := r.ctl_ieerror;
                when "001" =>   -- read status register
                    v_prdata(1 downto 0)    := r.sta_link;
                    v_prdata(2)     := r.sta_errdisc;
                    v_prdata(3)     := r.sta_errpar;
                    v_prdata(4)     := r.sta_erresc;
                    v_prdata(5)     := r.sta_errcred;
                    v_prdata(10)    := r.sta_gottick;
                    v_prdata(13)    := r.sta_rxpacket;
                    v_prdata(14)    := r.sta_rxempty;
                when "010" =>   -- read transmission clock scaler
                    v_prdata(7 downto 0)    := r.txdivcnt;
                    v_prdata(15 downto 8)   := r.tx_delay_data;
                    v_prdata(23 downto 16)  := r.tx_delay_strobe;
                when "011" =>   -- read timecode register
                    v_prdata(5 downto 0)    := linko.time_out;
                    v_prdata(7 downto 6)    := linko.ctrl_out;
                    v_prdata(13 downto 8)   := r.time_in;
                    v_prdata(16 downto 14)  := "000";
                when "111" =>   -- read version register
                    v_prdata(31 downto 24)  := SPW_VER;
                    v_prdata(23 downto 0)   := SPW_CFG;
                when others =>
                    null;
            end case;
        end if;

        -- Avalon-MM write access.
        if (avs_write) = '1' then
            case avs_address(2 downto 0) is
                when "000" =>   -- write control register
                    v.ctl_reset     := avs_writedata(0);
                    v.ctl_linkstart := avs_writedata(2);
                    v.ctl_autostart := avs_writedata(3);
                    v.ctl_linkdis   := avs_writedata(4);
                    v.ctl_ticken    := avs_writedata(5);
                    v.ctl_rxstart   := avs_writedata(6);
                    v.ctl_txstart   := avs_writedata(7);
                    if avs_writedata(8) = '1' then v.ctl_txcancel := '1'; end if;
                    v.ctl_ielink    := avs_writedata(9);
                    v.ctl_ietick    := avs_writedata(10);
                    v.ctl_ierxdesc  := avs_writedata(11);
                    v.ctl_ietxdesc  := avs_writedata(12);
                    v.ctl_ierxpacket := avs_writedata(13);
                    v.ctl_ieerror   := avs_writedata(14);
                when "001" =>   -- write status register
                    if avs_writedata(2) = '1' then v.sta_errdisc := '0'; end if;
                    if avs_writedata(3) = '1' then v.sta_errpar  := '0'; end if;
                    if avs_writedata(4) = '1' then v.sta_erresc  := '0'; end if;
                    if avs_writedata(5) = '1' then v.sta_errcred := '0'; end if;
                    if avs_writedata(10) = '1' then v.sta_gottick := '0'; end if;
                    if avs_writedata(13) = '1' then v.sta_rxpacket := '0'; end if;
                when "010" =>   -- write transmission clock scaler
                    v.txdivcnt      := avs_writedata(7 downto 0);
                    v.tx_delay_data := avs_writedata(15 downto 8);
                    v.tx_delay_strobe := avs_writedata(23 downto 16);
                when "011" =>   -- write timecode register
                    v.time_in       := avs_writedata(13 downto 8);
                    if avs_writedata(16) = '1' then v.tick_in := '1'; end if;
                when others =>
                    null;
            end case;
        end if;

        -- Drive control signals to RX fifo.
        s_rxfifo_raddr  <= v.rxfifo_raddr;  -- new value of rxfifo_raddr
--        s_rxfifo_wen    <= (not r.rxfull) and (linko.rxchar or r.rxeep);
        s_rxfifo_wen    <= r.rxfifo_write;
        s_rxfifo_waddr  <= r.rxfifo_waddr;
        s_rxfifo_wdata  <= r.rxfifo_wdata(9) & linko.rxflag & r.rxfifo_wdata(7 downto 0);

        -- Drive inputs to spwlink.
        linki.autostart <= r.ctl_autostart;
        linki.linkstart <= r.ctl_linkstart;
        linki.linkdis   <= r.ctl_linkdis;
        linki.rxroom    <= r.rxroom;
        linki.tick_in   <= r.tick_in;
        linki.ctrl_in   <= "00";
        linki.time_in   <= r.time_in;

        -- Drive divcnt input to spwxmit.
        if linko.running = '1' or r.tx_delay_data(6) = '1' then
            xmit_divcnt <= r.txdivcnt;
        else
            xmit_divcnt <= default_divcnt;
        end if;

	xmit_delay_data <= r.tx_delay_data;
	xmit_delay_strobe <= r.tx_delay_strobe;

        -- Drive outputs.
--        rxhalff         <= r.rxhalff;

	if (v.sta_rxempty = '0') then
            aso_valid   <= r.rxfifo_rvalid;
            aso_sop     <= s_rxfifo_rdata(9);
            aso_eop     <= s_rxfifo_rdata(8);
            aso_data    <= s_rxfifo_rdata(7 downto 0);
	else
            aso_valid   <= r.rxfifo_rvalid;
            aso_sop     <= '0';
            aso_eop     <= '0';
            aso_data    <= "00000000";
	end if;	

        -- Pass tick_out signal to output port.
        tick_out    <= linko.tick_out;

        -- Drive Avalon output signals.
        avs_readdata    <= v_prdata;
        avs_irq         <= r.irq;

        -- Reset components.
        link_rst        <= (not rstn) or r.ctl_reset;
        xmit_rst        <= not rstn;

        if r.ctl_txcancel = '1' then
            v.txpacket      := '0';
            v.txdiscard     := '0';
        end if;

        -- Reset registers.
        if rstn = '0' or r.ctl_reset = '1' then
            v   := regs_reset;
        end if;

        -- Update registers.
        rin <= v;
    end process;

    -- Update registers.
    process (clk) is
    begin
        if rising_edge(clk) then
            r <= rin;
        end if;
    end process;

end architecture spwavl_arch;
